#include <ext/hash_map>

#ifndef _BT_PEER_H_
#define _BT_PEER_H_

#define CHOKED 0x01
#define UNCHOKED 0x00

#define INTERESTED 0x01
#define NOT_INTERESTED 0x00

/* number of self initiated connections -- 60 secs*/
#define MAX_SELF_INITIATE 35 

#define MIN_PEERS 20

#define MAX_CONNECT 55

/* max time between the calculation of rate -- 50 secs */
#define MAX_RATE_PERIOD 50
/* default length of a piece ~ 250KB */ 
#define PIECE_LEN 256000

/* number of pieces to start rarest first from */
#define RAREST_FIRST_CUTOFF 1

/* maximum number of uploads allowed */
#define MAX_UPLOADS 5 
/* value to tell that there is no next piece ( should be > than max number of pieces */
#define NO_NEXT_PIECE 12111
/* time for which a node stays in the network after finishing its download */
#define DEFAULT_SEED_TIME 600

/* seconds between evaluating chokes*/
#define CHOKE_INTERVAL 10

/* seconds between rerequests */
#define REREQUEST_INTERVAL 600

/* seconds after startup to wait before unchoking */
#define INITIAL_CHOKE_INTERVAL 30

#define CLEAN_CHOKE 1
#define NO_CLEAN_CHOKE 2


#define NEW_PEER 0
#define EXISTING_PEER -1
#define MAX_CONNECTED_PEER -2

namespace stdext = ::__gnu_cxx;

class bt_peer;

class HashBtPeer
{
public:
   unsigned int operator()(bt_peer* p) const
   {
     //return stdext::hash<unsigned int>()((unsigned int)p);
	return (unsigned int)p;
   }
};

// This class' function operator() tests if any two keys are equal.

/* struct or */
class HashBtPeerCompare
{
 public:
   bool operator()(bt_peer* p1, bt_peer* p2) const
   {
     return p1 == p2;
   }
};

/* The peer class - has functions for all tasks a peer does */
class bt_peer{
	
public:
	/* unique id/index of this peer */
	unsigned int peer_id;

	/* number of pieces in the file being shared */
	unsigned int numpieces ;

	/* to determine how long this peer will stay after completing its download */
	unsigned int grace_seed_time;

	/* STATE VARIABLES */
	unsigned char* local_bitfield;
	unsigned int numgot;
	unsigned int numconnections;
	bool is_active_peer ;
	bool done; 
	bool peer_is_seed;	/* added on 05/15/05*/
	unsigned char rrq_count;	/* added on 05/15/05*/
	
	/****************************************************
         * stdext::hash_maps to find connection specific state info *
         ***************************************************/
		/*------ DOWNLOADER MAPS------- */
	/* stdext::hash_map to find whether the D/L from a peer is choked*/
	stdext::hash_map<bt_peer*,unsigned char,HashBtPeer, HashBtPeerCompare> dl_choked;
	
	/* stdext::hash_map to find whether interested in D/L from this peer*/
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>dl_interested;

	/* stdext::hash_map to find D/L rate from this peer */
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>dl_rate;
	
	/* stdext::hash_map to determine ratesince_time for D/L */
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>dl_ratesince_time;

	/* stdext::hash_map to find last time D/L rate calc*/
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>dl_last_time;

	/* stdext::hash_map to determine the curr_piece being D/L from a peer*/
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>curr_piece;

	/* stdext::hash_map to determine the size of the piece D/L */
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>curr_piece_downloaded;

		/*------UPLOADER MAPS -------------*/
	/* stdext::hash_map to find whether the U/L to this peer is choked*/
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>ul_choked;

	/* stdext::hash_map to find whether the U/L to this peer is interested*/
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>ul_interested;

	/* stdext::hash_map to find U/L rateseince_time */
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare >ul_ratesince_time;

	/* stdext::hash_map to find last time U/L rate calc*/
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>ul_last_time;
	
	/* stdext::hash_map to find U/L rate */
	stdext::hash_map<bt_peer*, unsigned int, HashBtPeer, HashBtPeerCompare>ul_rate;

	/***********CHOKER DATA*******************************************/

	/* list of current connections maintained by the choker*/
	vector<bt_peer*> choker_connections;
	unsigned int choke_count;

	/**************PIECE-PICKER DATA**********************************/

	/* array of number of peers having each piece (NEEDED by PiecePicker)*/
	unsigned int* numinterests;

	/* list of lists of pieces that are with i peers */
	vector< vector<unsigned int>* > interests;

	/* array of position of each piece in the corresponding list of 'interests' */
	unsigned int* pos_in_interests;

	/* called when a node is to join the network
	 * is_seed = TRUE if the node is a seed
	 * id = the ID of the joining node
	 * numpieces = numpieces in the file being shared */
	void join(bool is_seed, unsigned int id, unsigned int numpieces);

	/* called when a node is to leave a network*/
	void leave();	/* when this peer wants to leave on its own*/


	/* to check if I am currently idle */
	bool peer_is_idle();
	int num_downloads_curr();
	/* to log my current activity */
	void log_peer_info();

	void remove_downloader(bt_peer*);	/* when another peer is leaving -- to close the corresponding downloader*/
	void remove_uploader(bt_peer*);		/* when anoter peer is leaving -- to close the corresponding uploader */
	void close_all_connections();		/* when this peer is leaving on its own*/

	/* function to contact tracker*/
	void rerequest();

	/* function to establish a new connection with another peer
	 * needs: 1) @ param: pointer to other side peer */
	int start_connection(bt_peer* other_peer);

	/* connection made from the other side 
	 * needs: 1) @ pointer to the calling bt_peer object
		  					*/

	void external_connection_made(bt_peer* other_peer );

	/* function to do opt. unchoking */
	void choker();

	/* function to select which piece to download */
	stdext::hash_map<unsigned int, bool>pieces_in_progress;
	unsigned int piece_picker(unsigned char*);

	/* function receive notification from topology about bytes to receive*/
	/* params: 1) @param to the pointer to the bt_peer		     *
	 *       : 2) number of bytes to receive			     */
	void receive_bytes(bt_peer* sender_peer, unsigned int numbytes);
	
	void rechoke();
	void shift_over(unsigned int, vector<unsigned int> *, vector<unsigned int> *);

	/* transfer request from a gone downloader to another peer */
	void transfer_request(bt_peer*, unsigned int);

	/**************DOWNLOADER FUNCTIONS*********************************************/
	void download_startup(bt_peer*);

	/* function to send 'interested'*/
	void send_interested(bt_peer*);

	/* function to send 'not interested' */
	void send_not_interested(bt_peer*);

	/* function to send 'have' */
	void send_have(bt_peer*);

	/* function to receive 'have' */
	void got_have(bt_peer*, unsigned int);

	/* function to request another piece*/
	void request_more(bt_peer*);

	/* function to receive the data from the network
	 * params: 1) number of bytes received */
	void got_data(bt_peer*, unsigned int);

	/* function to receive a piece */
	void got_piece(bt_peer*, unsigned int);

	/* function to receive a choke message */
	void got_choke(bt_peer*);

	/* function to receive an unchoke message */
	void got_unchoke(bt_peer*);

	/* function to update download rate */
	void download_update_rate(bt_peer*, unsigned int amount);

	/* cleans up state associated with the downloader*/
	void download_cleanup_state(bt_peer*, unsigned char);
	/* called up by the other side */
	void download_close_connection(bt_peer*);
	void check_min_connections();


	/*******************UPLOADER FUNCTIONS*****************************************/
	/* function to reply to a request*/
	void got_request(bt_peer*);

	/* function to receive a cancel message */
	void got_cancel(bt_peer*);

	/* function to send a choke message */
	void send_choke(bt_peer*);

	/* function to send an unchoke message */
	void send_unchoke(bt_peer*);

	/* sends a have message to the connected downloaders*/
	void send_have(bt_peer*, unsigned int);

	/* function to receive an 'interested' message */
	void got_interested(bt_peer*);

	/* function to receive a 'not interested' message */
	void got_not_interested(bt_peer*);
	
	/* function to update upload rate*/
	void upload_update_rate(bt_peer*, unsigned int);

	/* cleans up state associated with the downloader*/
	void upload_cleanup_state(bt_peer*);
	/* called up by the other side */
	void upload_close_connection(bt_peer*);

};

#endif /* _BT_PEER_H_ */
