#include <assert.h>
#include "memory_pool.h"
#include "util.h"

/**********
 * MACROS *
 **********/

#define ASSERT_POOL(pool) \
	do { \
		assert(pool); \
		assert(pool->pool_objects); \
		assert(pool->num_objects > 0); \
		assert(pool->used_bitmap); \
		assert(pool->num_words > 0); \
	} while (0)

/*************
 * CONSTANTS *
 *************/


static const unsigned int BITS_PER_BYTE = 8;
static const unsigned int OBJECTS_PER_WORD =
	BITS_PER_BYTE * sizeof(unsigned int);


/************************
 * FUNCTION DEFINITIONS *
 ************************/


void init_memory_pool(struct memory_pool *pool,
		unsigned int num_objects,
		size_t obj_size)
{
	assert(pool && num_objects > 0);
	memset(pool, 0, sizeof(*pool));

	pool->obj_size = obj_size;

	/* round objects up to OBJECTS_PER_WORD */
	unsigned int word_offset = num_objects % OBJECTS_PER_WORD;
	pool->num_objects = num_objects;
	if (word_offset > 0) {
		pool->num_objects += OBJECTS_PER_WORD - word_offset;
	}

	pool->pool_objects = _malloc(pool->num_objects * pool->obj_size);

	pool->num_words = pool->num_objects / OBJECTS_PER_WORD;
	pool->used_bitmap = (unsigned int *)_malloc(pool->num_words *
			sizeof(unsigned int));

	pool->next_word_offset = 0;
}

void *pool_object_alloc(struct memory_pool *pool)
{
	ASSERT_POOL(pool);
	unsigned int start_offset = pool->next_word_offset;

	do {
		unsigned int word_value =
			pool->used_bitmap[pool->next_word_offset];
		if (word_value == 0xFFFFFFFF) {
			pool->next_word_offset++;
			pool->next_word_offset =
				pool->next_word_offset % pool->num_words;
			continue;
		}

		unsigned int object_offset =
			pool->next_word_offset * OBJECTS_PER_WORD;
		unsigned int tmp = 1;

		while (tmp) {
			if (!(word_value & tmp)) {
				char *ptr = (char*)pool->pool_objects;
				ptr += object_offset * pool->obj_size;
				pool->used_bitmap[pool->next_word_offset] |=
					tmp;
				return (void*)ptr;
			}
			object_offset++;
			tmp <<= 1;
		}

		/* should never reach here */
		assert(0);
	} while (start_offset != pool->next_word_offset);

	return NULL;
}

void pool_object_free(struct memory_pool *pool, void *obj_ptr)
{
	assert(obj_ptr);
	ASSERT_POOL(pool);

	unsigned int object_offset = ((char*)obj_ptr -
			(char*)pool->pool_objects) / pool->obj_size;
	unsigned int word_offset = object_offset / OBJECTS_PER_WORD;
	unsigned int bit_offset = object_offset % OBJECTS_PER_WORD;
	unsigned int tmp = 1 << bit_offset;

	pool->used_bitmap[word_offset] &= ~tmp;
}
