#ifndef _TOPOLOGY_H_
#define _TOPOLOGY_H_

#include "btsim.h"

enum {
	INTRA_HEADEND = 0,
	INTRA_UNIV,
	INTRA_AS,
	INTER
};

enum {
	NORMAL_TRACKER = 0,
	SUB_TRACKER,
	UNIV_TRACKER,
	SUB_LIMIT_TRACKER	/* added by Ruchir to support per AS conn. limit trackers */
};


/*************
 * CONSTANTS *
 *************/


const unsigned char DEGREES_OF_CLOSENESS = 4;


/********************
 * GLOBAL VARIABLES *
 ********************/

extern unsigned long long g_closeness[DEGREES_OF_CLOSENESS];


/***********************
 * FUNCTION PROTOTYPES *
 ***********************/

class bt_peer;


/**
 * Initializes the topology by reading its description from a file.
 * @param topo_file	the name of the topology file
 * @return 0 on success, non-zero on error.
 */
int init_topology(const char *topo_file);

/**
 * Returns the number of bit torrent peers in the topology that are university
 * nodes.
 * @return the number of university nodes
 */
unsigned int get_num_univ_nodes();

/**
 * Returns the number of bit torrent peers in the topology, both active and 
 * inactive.
 * @return the number of bit torrent peers in the topology.
 */
unsigned int get_num_peers();

/**
 * Returns the array of bit torrent peers in the topology.  Note this includes 
 * both active and inactive peers.
 * @return the bt_peer array.
 */
bt_peer *get_peers();

class bt_tracker;

/**
 * Query the topology to obtain a pointer to the tracker object.
 * @param peer		the bit torrent peer querying the tracker
 * @return the tracker / sub_tracker object
 */
bt_tracker *get_tracker(bt_peer *peer);

/**
 * Adds a one way data transfer to the topology.  This initializes a data 
 * transfer for shared link calculations and calling back to the receiver
 * node.
 * @param sender	the sending bt_peer
 * @param receiver	the receiving bt_peer, will get callbacks
 */
void add_data_transfer(bt_peer *sender,
		bt_peer *receiver);

/**
 * Removes a one way data transfer from the topology.
 * @param sender	the sending bt_peer
 * @param receiver	the receiving bt_peer, will get callbacks
 */
void remove_data_transfer(bt_peer *sender, bt_peer *receiver);

/**
 * Calculates progress made on data transfers.
 * @param tick		the current time to update transfer to
 */
void calc_data_transfers(unsigned int tick);

/**
 * Returns whether the seed is a university node or a cable leaf node.
 * @return 0 if the seed is a university node, non-zero if a cable leaf node
 */
int get_seed_is_cable();

/**
 * Returns whether the peer arrivals are a flash crowd (all at once) or a 
 * poisson process.
 * @return 0 if the arrivals are a flash crowd, non-zero if a poisson process
 */
int get_arrivals_are_random();

/**
 * Returns the number of pieces that compose the file being downloaded.
 * @return the number of pieces
 */
unsigned int get_num_pieces();

/**
 * Returns whether the departures are immediately after the download or 
 * a poisson process after the download completes.
 * @return 0 if the peers depart immediately after the download completes,
 * non-zero if the departures are Poisson after the download completes.
 */
int get_departures_are_random();

/**
 * Accessor for the tracker type.
 * @return the tracker type
 */
unsigned int get_tracker_type();

/**
 * Accessor for subtracker / univ tracker option.
 * @return the number of peers outside of AS for subtracker or the
 * number of non university nodes for univ biased tracker
 */
unsigned int get_num_external_peers();

/**
 * Logs the bandwidth usage information.
 */
void log_bw_usage();

#endif /* _TOPOLOGY_H_ */
