#ifndef _UTIL_H_
#define _UTIL_H_

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "btsim.h"

/**********
 * MACROS *
 **********/

/* used for unreachable paths (if they're reachable, it's a bug) */
#define BUG() assert(0)

/* debugging / logging macros */

#define DEBUG_ERROR(fmt, args...) \
	do { \
		debug_msg("(%s:%d) Time: %d - ", \
				__FILE__, \
				__LINE__, \
				g_tick); \
		debug_msg(fmt, ## args); \
	} while (0)

#define DEBUG_MSG(fmt, args...) \
	do { \
		debug_msg("Time: %d - ", \
				g_tick); \
		debug_msg(fmt, ## args); \
	} while (0)

#define LOG_MSG(fmt, args...) \
	do { \
		log_msg("Time: %d - ", \
				g_tick); \
		log_msg(fmt, ## args); \
	} while (0)

/* safe cleanup macros */

#define SAFE_FREE(ptr) \
	do { \
		if (ptr) { \
			free(ptr); \
			ptr = NULL; \
		} \
	} while (0)

#define SAFE_DELETE(ptr) \
	do { \
		if (ptr) { \
			delete (ptr); \
			ptr = NULL; \
		} \
	} while (0)

#define SAFE_CLOSE(fp) \
	do { \
		if (fp) { \
			fclose(fp); \
			fp = NULL; \
		} \
	} while (0)

/* expect macros, expect a certain value, else set return value and goto 
   _cleanup label */

	/* synonyms */

#define EXPECT_ZERO		EXPECT_FALSE
#define EXPECT_NONZERO		EXPECT_TRUE
#define EXPECT_NOTNULL		EXPECT_TRUE

#define _EXPECT_ZERO		_EXPECT_FALSE
#define _EXPECT_NONZERO		_EXPECT_TRUE
#define _EXPECT_NOTNULL		_EXPECT_TRUE

#define EXPECT_FALSE(x) \
	do { \
		if (x) { \
			ret = -1; \
			goto _cleanup; \
		} \
	} while (0)

#define _EXPECT_FALSE(x, fmt, args...) \
	do { \
		if (x) { \
			DEBUG_ERROR(fmt, ## args); \
			ret = -1; \
			goto _cleanup; \
		} \
	} while (0)

#define EXPECT_TRUE(x) \
	do { \
		if (!(x)) { \
			ret = -1; \
			goto _cleanup; \
		} \
	} while (0)

#define _EXPECT_TRUE(x, fmt, args...) \
	do { \
		if (!(x)) { \
			DEBUG_ERROR(fmt, ## args); \
			ret = -1; \
			goto _cleanup; \
		} \
	} while (0)


/***********************
 * FUNCTION PROTOTYPES *
 ***********************/

/**
 * Enables debugging. Sets the global debugging flag.
 */
void debugging_enable();

/**
 * Disables debugging. Unsets the global debugging flag.
 */
void debugging_disable();

/**
 * Prints a debug msg to stderr
 * Checks to see if the debug flag is on, if so, prints the debug message.
 * @param format	the format string
 * @param ...		the variable argument list
 */
void debug_msg(const char *format, ...);

/**
 * Sets the logfile name.  Tries to open the file.
 * @param logname	the logfile name
 * @return 0 on success, non-zero on error.
 */
int open_logfile(const char *logname);

/**
 * Closes the logfile.  Nothing happens if there is no open logfile.
 */
void close_logfile();

/**
 * Enables logging.  Sets the global logging flag.
 */
void logging_enable();

/**
 * Disables logging.  Unsets the global logging flag.
 */
void logging_disable();

/**
 * Logs a message to the logfile.
 * Checks to see if the logging flag is on, if so, logs the message.
 * @param format	the format string
 * @param ...		the variable argument list
 */
void log_msg(const char *format, ...);

/**
 * Mallocs memory.  Asserts on error.  Initializes memory to 0.
 * @param size		amount of memory to allocate
 * @return a pointer to the memory
 */
static __inline__ void *_malloc(size_t size)
{
	void *ptr = malloc(size);
	assert(ptr);
	memset(ptr, 0, size);
	return ptr;
}

/**
 * Reallocates memory.  Asserts on error.
 * @param size		amount of memory to allocate
 * @return a pointer to the memory
 */
static __inline__ void *_realloc(void *ptr, size_t size)
{
	void *tmp = realloc(ptr, size);
	assert(tmp);
	return tmp;
}

#endif /* _UTIL_H_ */
